/**
 *  Matrix.h
 *
 *  Mtrix osztly
 *
 *  Mtrix mveletek megvalstsa, invertls gauss elimincival,
 *  diagonlizci forgatmtrixokkal, SVD felbonts, SVD alapjn pszeud
 *  inverz, kivetel kezeles, es Vektor megvalositas
 *  
 */

#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <stdarg.h>
#include <math.h>

#include <iostream>
#include <string>


/**
 *  Matrix muveletek hibajanak (pl.: meretek nem egyeznek)
 *  jelzesere irt kivetelosztaly
 */
class MatrixException {
        protected:
                std::string message;
        public:
                MatrixException( std::string message ) {
                        this->message = message;
                }
        public:
                friend std::ostream& operator << ( std::ostream& o, MatrixException& e ) {
                        o << e.message;

                        return o;
                }
};

/**
 *  Mtrix inverz szmtsnak tipusa
 */
typedef enum {
        GAUSS
} InverseMethod;

/**
 *  Matrix osztaly
 */
template<class ElemType = float>
class Matrix {
        public:
                int n, m;
                typedef ElemType * ElemP;
                ElemP * elems;

        public:
                /** Konstruktor */
                Matrix( ) {
                        elems = 0;
                        n = -1; m = -1;
                }

                /** Konstruktor */
                Matrix( int n, int m ) {
                        elems = 0;
                        alloc( n, m );
                }

                /** Destructor */
                ~Matrix( ) {
                        free( );
                }
                
                /** Konstruktor */
                Matrix(const Matrix<ElemType>& mat ) {
                        int i, j;

                        elems = 0;
                        alloc( mat.n, mat.m );

                        for( i = 0; i < mat.n; i++)
                                for( j = 0; j < mat.m; j++ )
                                        getElement( i, j ) = mat.elems[i][j];
                }

                Matrix<ElemType>& operator=(const Matrix<ElemType>& mat ) {
                        int i, j;

                        if ( n != mat.n || m != mat.m )
                                alloc( mat.n, mat.m );

                        for( i = 0; i < mat.n; i++)
                                for( j = 0; j < mat.m; j++ )
                                        getElement( i, j ) = mat.elems[i][j];
                        return *this;
                }

                inline ElemType& getElement( int i, int j ) {
                        return elems[i][j];
                }

                int getN( void ) {
                        return n;
                }

                int getM( void ) {
                        return m;
                }
                
                void setSize( int n, int m ) {
                        this->n = n;
                        this->m = m;
                }

                friend std::ostream& operator << ( std::ostream& o, Matrix<ElemType>& matrix ) {
                        int i, j;

                        for( i = 0; i < matrix.n; i++) {
                                for( j = 0; j < matrix.m; j++ ) {
                                        if ( fabs(matrix.elems[i][j]) < 0.0001f )
                                                o << "0 ";
                                        else
                                                o << matrix.elems[i][j] << " ";
                                }
                                o << std::endl;
                        }

                        return o;
                }

                Matrix<ElemType> operator | ( const Matrix<ElemType>& m2 ) {
                        if( n != m2.n )
                                throw MatrixException( "" );

                        Matrix<ElemType> ret( n, m+m2.m );
                        int i, j;

                        for( i = 0; i < n; i++) {
                                for( j = 0; j < m; j++ )
                                        ret.elems[i][j] = elems[i][j];
                                for( j = 0; j < m2.m; j++ )
                                        ret.elems[i][m+j] = m2.elems[i][j];
                        }

                        return ret;
                }
                Matrix<ElemType> T( void ) {
                        Matrix<ElemType> ret( m, n );
                        int i, j;
                        for( i = 0; i < n; i++)
                                for( j = 0; j < m; j++ )
                                        ret.elems[j][i] = elems[i][j];
                        return ret;
                }

                inline friend void fastTranspMatrix( const Matrix<ElemType>& m, Matrix<ElemType>& dst ) {
                        if ( dst.n != m.m || dst.m != m.n )
                                dst.alloc( m.m, m.n );

                        int i, j;

                        for( i = 0; i < m.m; i++ )
                                for( j = 0; j < m.n; j++ )
                                        dst.elems[i][j] = m.elems[j][i];
                }

                Matrix<ElemType> operator*( const Matrix<ElemType>& mat ) {
                        if( m != mat.n )
                                throw MatrixException( "Nem egyeznek a mretek!" );

                        Matrix<ElemType> ret(n, mat.m);

                        int i, j, t;

                        for( i = 0; i < n; i++ )
                                for( j = 0; j < mat.m; j++ ) {
                                        ElemType sum = 0;
                                        for( t = 0; t < m; t++ )
                                                sum = sum + (elems[i][t]*mat.elems[t][j]);
                                        ret.elems[i][j] = sum;
                                }

                        return ret;
                }

                inline friend void fastMultMatrix( const Matrix<ElemType>& m1, const Matrix<ElemType>& m2, Matrix<ElemType>& dst ) {
                        if( m1.m != m2.n )
                                throw MatrixException( "" );

                        if ( dst.n != m1.n || dst.m != m2.m )
                                dst.alloc( m1.n, m2.m );

                        int i, j, t;

                        for( i = 0; i < m1.n; i++ )
                                for( j = 0; j < m2.m; j++ ) {
                                        ElemType sum = 0;
                                        for( t = 0; t < m1.m; t++ )
                                                sum = sum + (m1.elems[i][t]*m2.elems[t][j]);
                                        dst.elems[i][j] = sum;
                                }
                }

                Matrix<ElemType> operator-( const Matrix<ElemType>& mat ) {
                        if( m != mat.m || n !=mat.n )
                                throw MatrixException( "" );

                        Matrix<ElemType> ret(n, m);
                        int i, j;

                        for( i = 0; i < n; i++ )
                                for( j = 0; j < m; j++ )
                                        ret.elems[i][j] = elems[i][j]-mat.elems[i][j];

                        return ret;
                }

                inline friend void fastSubMatrix( const Matrix<ElemType>& m1, const Matrix<ElemType>& m2, Matrix<ElemType>& dst ) {
                        if( m1.m != m2.m || m1.n !=m2.n )
                                throw MatrixException( "" );

                        if ( dst.n != m1.n || dst.m != m1.m )
                                dst.alloc( m1.n, m1.m );

                        int i, j, t;

                        for( i = 0; i < m1.n; i++ )
                                for( j = 0; j < m1.m; j++ )
                                        dst.elems[i][j] = m1.elems[i][j]-m2.elems[i][j];
                }


                Matrix<ElemType> inverse( InverseMethod method = GAUSS ) {
                        if ( n != m )
                                throw MatrixException( "Only square matrices can be inverted!" );

                        switch( method ) {
                                case GAUSS: return inverseGauss( );
                        }
                        throw MatrixException( "Unknown inverse method type!" );
                }

                inline friend void  fastInverseMatrix( Matrix<ElemType>& m, Matrix<ElemType>& dst/*, InverseMethod method = GAUSS */) {
                        if ( m.n != m.m )
                                throw MatrixException( "Only square matrices can be inverted!" );

                        fastInverseGauss( m, dst );
/*                         switch( method ) {
                                case GAUSS: fastInverseGauss( m, dst ); return;
                        }
                        throw MatrixException( "Unknown inverse method type!" );*/
                }

                Matrix<ElemType> inverseGauss( void ) {
                        int i, j, k;
                        ElemType e, temp;

                        Matrix<ElemType> A(n, n*2);
                        Matrix<ElemType> E(n, n);

                        A = ((*this) | E );

                        for( i = 0; i < n; i++ ) {
                                if( fabs(A.elems[i][i]) <= 0.01f ) {
                                        for( j = 0; j < n; j++ ) {
                                                if( i == j ) continue;
                                                if( fabs(A.getElement(j, i)) >= 0.01f ) {
                                                        for( int k = 0; k < 2*n; k++ )
                                                        A.elems[i][k] += A.elems[j][k];
                                                        break;
                                                }
                                        }
                                        if( j == n ) {
                                                throw MatrixException( "Matrix Error" );
                                        }
                                }
                        }


                        for( i = 0; i < n; i++ ) {
                                if( fabs(A.elems[i][i]) <= 0.01f ) {
                                        for( j = 0; j < n; j++ ) {
                                                if( i == j ) continue;
                                                for( k = 0; k < i; k++ )
                                                        if( A.getElement( j, k ) != 0.0f )
                                                                break;
                                                if( k != i ) continue;
                                                if( fabs(A.getElement(j, i)) >= 0.01f ) {
                                                        for( int k = 0; k < 2*n; k++ )
                                                                A.elems[i][k] += A.elems[j][k];
                                                        break;
                                                }
                                        }
                                        if( j == n ) {
                                                throw MatrixException( "Matrix Error" );
                                        }
                                }

                                e = A.getElement( i , i );

                                for( j = 0; j < 2*n; j++ )
                                        A.elems[i][j] /= e;
                                for( k = 0; k < n; k++ ) {
                                        if ( k == i )
                                                continue;

                                        temp = A.elems[k][i] / A.elems[i][i];

                                        for( j = 0; j < 2*n; j++ )
                                                A.elems[k][j] -= A.elems[i][j]* temp;
                                }

                        }

                        for( i = 0; i < n; i++ )
                                for( j = 0; j < n; j++ )
                                        E.elems[i][j] = A.elems[i][n+j];

                        return E;
                }

                inline friend void fastInverseGauss( Matrix<ElemType>&m, Matrix<ElemType>& dst ) {
                        int i, j, k;
                        int n = m.n;
                        ElemType e, temp;
                        ElemType C[n][n];

                        if ( dst.n != n || dst.m != n ) {
                                dst.alloc( m.n, m.n );
                        }
                        for( i = 0; i < n; i++ )
                                for( j = 0; j < n; j++ ) {
                                        dst.elems[i][j] = (i==j)?(ElemType)1:(ElemType)0;
                                        C[i][j] = m.elems[i][j];
                                }

                        for( i = 0; i < n; i++ ) {
                                if( fabs(C[i][i]) <= 0.01f ) {
                                        for( j = 0; j < n; j++ ) {
                                                if( i == j ) continue;
                                                if( fabs( C[j][i] ) >= 0.01f ) {
                                                        for( int k = 0; k < n; k++ ) {
                                                                C[i][k] += C[j][k];
                                                                dst.elems[i][k] = dst.elems[j][k];
                                                        }
                                                        break;
                                                }
                                        }
                                        if( j == n ) {
                                                throw MatrixException( "Matrix Error" );
                                        }
                                }
                        }


                        for( i = 0; i < n; i++ ) {
                                e = C[i][i];

                                for( j = 0; j < n; j++ ) {
                                        C[i][j] /= e;
                                        dst.elems[i][j] /= e;
                                }
                                for( k = 0; k < n; k++ ) {
                                        if ( k == i )
                                                continue;

                                        temp = C[k][i] / C[i][i];

                                        for( j = 0; j < n; j++ ) {
                                                C[k][j] -= C[i][j]* temp;
                                                dst.elems[k][j] -= dst.elems[i][j]* temp;
                                        }
                                }
                        }
                }

                friend void createRotationMatrix( Matrix<ElemType>& M, int n, int i, int j, ElemType alpha ) {
                        int ii, jj;
                        if ( M.n != n || M.m != n )
                                M.alloc( n, n );

                        for( ii = 0; ii < n; ii++ ) {
                                for( jj = 0; jj < n; jj++ )
                                        M.elems[ii][jj] = (ii==jj)?(ElemType)1:(ElemType)0;
                        }

                        M.elems[i][i] =  cos( alpha );
                        M.elems[i][j] = -sin( alpha );
                        M.elems[j][i] =  sin( alpha );
                        M.elems[j][j] =  cos( alpha );
                }

        friend void jacobiDecomposition( Matrix<ElemType>& A, Matrix<ElemType>& D, Matrix<ElemType>&Q ) {
                Matrix<ElemType> Ainv;
                int N = A.getN( );
                Matrix<ElemType> temp;
                Matrix<ElemType> J;
                Matrix<ElemType> Jtrans;
                int i, j, iter;
                ElemType Threshold, Maximum;
                ElemType alpha;

                D = A;
                Q = Matrix<ElemType>( N , N );

                Maximum = fabs(A.getElement(0, 0));
                for( i = 0; i < N-1; i++ )
                        for( j = i+1; j < N; j++ ) {
                                if ( fabs(A.getElement(i, j)) > Maximum )
                                        Maximum = fabs(A.getElement(i, j));
                        }

                Threshold = Maximum;


  //              for( iter=0; iter < 100; iter++ ) {
                while( 1 ) {
                        ElemType max;
                        int maxi, maxj;

                        maxi = 0; maxj = 1;
                        max = D.getElement( 0, 1 );
                        for( i = 0; i < N-1; i++ )
                                for( j = i+1; j < N; j++ ) {
                                        if ( max <fabs(D.getElement(i, j )) ) {
                                                max = fabs(D.getElement(i, j ));
                                                maxi = i;
                                                maxj = j;
                                        }
                                }
                        if ( max < 0.001 ) break;
                        i = maxi;
                        j = maxj;

                          alpha = atan2( D.getElement(i, j)*2.0f, (D.getElement(i, i) - D.getElement(j, j)  ) ) / 2.0f;

                          createRotationMatrix( J, N, i, j, alpha );
                          fastTranspMatrix( J, Jtrans );

                          fastMultMatrix( Jtrans, D, temp );
                          fastMultMatrix( temp, J, D );

                          fastMultMatrix( Q, J, temp );
                          Q = temp;
/*                        for( i = 0; i < N-1; i++ )
                                for( j = i+1; j < N; j++ )
                                        if( D.getElement(i, j) >= Threshold || D.getElement(i, j) <= -Threshold ) {
                                                alpha = atan2( D.getElement(i, j)*2.0f, (D.getElement(i, i) - D.getElement(j, j)  ) ) / 2.0f;

                                                createRotationMatrix( J, N, i, j, alpha );
                                                fastTranspMatrix( J, Jtrans );

                                                fastMultMatrix( Jtrans, D, temp );
                                                fastMultMatrix( temp, J, D );

                                                fastMultMatrix( Q, J, temp );
                                                Q = temp;
                                        }
                        Threshold /= 2.0f;*/
                }
        }

        friend void invDiagonal( Matrix<ElemType>& M, Matrix<ElemType>& dst ) {
                int i;
                int r = (M.getN() < M.getM())? M.getN(): M.getM();

                dst = M;

                for( i = 0; i < r; i++ )
                        if ( dst.elems[i][i] )
                                dst.elems[i][i] = ((ElemType)1) / dst.elems[i][i];
        }

        friend void SVDDecomposition( Matrix<ElemType> &M, Matrix<ElemType>& U, Matrix<ElemType>& S, Matrix<ElemType>& V ) {

                int i, j, k, r;
                Matrix<ElemType> A, D, Q;

                S = Matrix<ElemType>( M.getN(), M.getM() );
                Matrix<ElemType> Sinv = Matrix<ElemType>( M.getN(), M.getM() );
                if ( M.getN() <= M.getM() ) {
                        A = M*M.T();
                        jacobiDecomposition( A, D, Q );

                        for( j = M.getN()-1; j > 0; --j ) {
                                int mini = 0;
                                for( i = 1; i <= j; i++ ) {
                                        if ( D.elems[i][i] < D.elems[mini][mini] )
                                                mini = i;
                                }
                                ElemType temp;
                                temp = D.elems[mini][mini];
                                D.elems[mini][mini] = D.elems[j][j];
                                D.elems[j][j] = temp;
                                for( k = 0; k < U.getN(); k++ ) {
                                        temp = Q.elems[k][mini];
                                        Q.elems[k][mini] = Q.elems[k][j];
                                        Q.elems[k][j] = temp;
                                }
                        }
                        U = Q;

                        for( i = 0; i < M.getN(); i++) {
                                S.elems[i][i] = sqrt( D.elems[i][i] );

                                if( fabs( S.elems[i][i] ) > 0.0001 )
                                        Sinv.elems[i][i] = 1.0f / S.elems[i][i];
                                else
                                        Sinv.elems[i][i] = 0.0f;
                        }

                        V = (Sinv.T() * U.T() * M).T();

                        for( r = 0; r < S.getN(); r++ ) {
                                if( !S.elems[r][r] )
                                        break;
                        }

                        if( r < S.getN() )
                                r++;

                        for( i = r; i < V.getM( ); i++ ) {
                                for( j = 0; j < V.getM( ); j++ )
                                        V.elems[j][i] = 0.0f;
                                V.elems[0][i] = 1.0f;

                                for( k = 0; k < i; k++ ) {
                                        for( j = 0; j < V.getN( ); j++ )
                                                V.elems[j][i] -= V.elems[j][k] * V.elems[k][k] ;
                                }
                                ElemType length = (ElemType)0;
                                for( j = 0; j < V.getM( ); j++ )
                                       length += V.elems[j][i] * V.elems[j][i];
                                length = sqrt( length );
                                for( j = 0; j < V.getM( ); j++ )
                                       V.elems[j][i] /= length;
                        }
                } else {
                        A = M.T()*M;
                        jacobiDecomposition( A, D, Q );

                        for( j = M.getM()-1; j > 0; --j ) {
                                int mini = 0;
                                for( i = 1; i <= j; i++ ) {
                                        if ( D.elems[i][i] < D.elems[mini][mini] )
                                                mini = i;
                                }
                                ElemType temp;
                                temp = D.elems[mini][mini];
                                D.elems[mini][mini] = D.elems[j][j];
                                D.elems[j][j] = temp;
                                for( k = 0; k < U.getM(); k++ ) {
                                        temp = Q.elems[k][mini];
                                        Q.elems[k][mini] = Q.elems[k][j];
                                        Q.elems[k][j] = temp;
                                }
                        }
                        V = Q;

                        for( i = 0; i < M.getM(); i++) {
                                S.elems[i][i] = sqrt( D.elems[i][i] );

                                if( fabs( S.elems[i][i] ) > 0.0001 )
                                        Sinv.elems[i][i] = 1.0f / S.elems[i][i];
                                else
                                        Sinv.elems[i][i] = 0.0f;
                        }

                        U = (M*V * Sinv.T() );

                        for( r = 0; r < S.getM(); r++ ) {
                                if( !S.elems[r][r] )
                                        break;
                        }

                        if( r < S.getM() )
                                r++;

/*                        for( i = r; i < U.getN( ); i++ ) {
                                for( j = 0; j < U.getN( ); j++ )
                                        U.elems[j][i] = 0.0f;
                                U.elems[0][i] = 1.0f;

                                for( k = 0; k < i; k++ ) {
                                        for( j = 0; j < V.getN( ); j++ )
                                                U.elems[j][i] -= U.elems[j][k] * V.elems[k][k] ;
                                }
                                ElemType length = (ElemType)0;
                                for( j = 0; j < U.getM( ); j++ )
                                       length += U.elems[j][i] * U.elems[j][i];
                                length = sqrt( length );
                                for( j = 0; j < U.getM( ); j++ )
                                       V.elems[j][i] /= length;
                        }*/
                }

        }

       
        friend void SVDInverse( Matrix<ElemType>& M, Matrix<ElemType>&dst ) {
                Matrix<ElemType> U, S, V;

                SVDDecomposition( M , U, S, V );
                Matrix<ElemType> Sinv;
                invDiagonal( S, Sinv );

                Matrix<ElemType> USVT;
                USVT= U*S*V.T();

                dst = (U * Sinv *V.T()).T();
        }

        protected:
                void free( void ) {
                        int i;
                        if( elems ) {
                                delete elems[0];
                                delete elems;
                        }
                }
                void alloc( int n, int m ) {
                        int i, j;
                        free( );


                        elems = new ElemP[n];
                        elems[0] = new ElemType[n*m];
                        for( i = 0; i < n; i++ ) {
                                elems[i] = &elems[0][m*i];
                                for( j = 0; j < m; j++ )
                                        elems[i][j] = (i==j)?(ElemType)1:(ElemType)0;
                        }
                        this->n = n;
                        this->m = m;
                }

};

/**
 *  Vektor osztly a Mtrix-bol szarmaztatva
 *  A szukseges muveletek vannak csak atirva
 */
template<class ElemType>
class Vector : public Matrix<ElemType> {
        public:
                Vector( ) : Matrix<ElemType>( ) {
                }
                Vector( int n ) : Matrix<ElemType>( n , 1 ) {
                }

                void setSize( int n ) {
                        this->n = n;
                }

                Vector<ElemType>& operator = (  Matrix<ElemType>& m ) {
                        int i, j;

                        if ( m.getM() != 1 ) throw MatrixException( "Hiba" );

                        alloc( m.getN(), 1 );

                        Matrix<ElemType>::n = m.getN();
                        Matrix<ElemType>::m = 1;

                        for( i = 0; i < m.getN(); i++)
                                getElement( i ) = m.getElement( i, 0 );

                        return *this;
                }

                Vector<ElemType>& operator = ( Matrix<ElemType> m ) {
                        int i, j;

                        if ( m.getM() != 1 ) throw MatrixException( "Hiba" );

                        alloc( m.getN(), 1 );

                        Matrix<ElemType>::n = m.getN();
                        Matrix<ElemType>::m = 1;

                        for( i = 0; i < m.getN(); i++)
                                getElement( i ) = m.getElement( i, 0 );

                        return *this;
                }

                inline ElemType& getElement( int i ) {
                        return Matrix<ElemType>::elems[i][0];
                }

                ElemType length( void ) {
                        int i;
                        ElemType sum = (ElemType)0.0;
                        for( i = 0; i < this->n; i++)
                                sum += getElement( i ) * getElement( i );
                        return sqrt( sum );
                }
};

#endif /*  _MATRIX_H_ */
