#include "WebcamInterface.h"

/**
 *  OpenCV webcamera felvetel inditasa 
 */
WebcamInterface::WebcamInterface( int camID ) {
        int i;
        
        cvCapture = cvCaptureFromCAM( camID );

        image = 0;

        of_flags = 0;

        std::vector<float> motion;

        motion.push_back( 0 );
        motion.push_back( 0 );
        motion.push_back( 1 );
        motion.push_back( 0 );

        for( i = 0; i < MOTION_BUFFER_SIZE; i++ )
                shifts.push_back( motion );
}

/**
 *  OpenCV AVI felvetel inditasa 
 */
WebcamInterface::WebcamInterface( char * aviFileName ) {
        int i;
        
        cvCapture = cvCaptureFromAVI( aviFileName );

        image = 0;

        of_flags = 0;

        std::vector<float> motion;

        motion.push_back( 0 );
        motion.push_back( 0 );
        motion.push_back( 1 );
        motion.push_back( 0 );

        for( i = 0; i < MOTION_BUFFER_SIZE; i++ )
                shifts.push_back( motion );
}

/**
 *  Dinamikus valtozok felszabaditasa
 */
WebcamInterface::~WebcamInterface( ) {
        cvReleaseCapture( &cvCapture );

        if ( image ) {
                cvReleaseImage( &eig );
                cvReleaseImage( &temp );

                cvReleaseImage( &newImage );
                cvReleaseImage( &pyramid );
                cvReleaseImage( &newPyramid );

                delete points;
                delete newPoints;
                delete status;
        }
}

/**
 *  Dinamikus valtozok lefoglalasa, elso kep felvetele
 */
void WebcamInterface::init( void ) {
        IplImage * frame;

        if( !cvGrabFrame( cvCapture ) ) {
                std::cout << "Nem sikerult olvasni a frame-t!" << std::endl;
                exit(0);
        }
        frame = cvRetrieveFrame( cvCapture );

        /**
         *  Kpek, ideiglenes kpek, piramis kpek lefoglalsa
         */
        if( !image ) {
                grayFrame = cvCreateImage( cvGetSize(frame), 8, 1 );
                grayFrame->origin = frame->origin;

                image = cvCreateImage( cvSize( FRAME_WIDTH, FRAME_HEIGHT ), 8, 1 );
                image->origin = frame->origin;

                eig = cvCreateImage( cvGetSize(image), 32, 1 );
                temp = cvCreateImage( cvGetSize(image), 32, 1 );

                newImage = cvCreateImage( cvGetSize(image), 8, 1 );
                pyramid = cvCreateImage( cvGetSize(image), 8, 1 );
                newPyramid = cvCreateImage( cvGetSize(image), 8, 1 );
                points = new CvPoint2D32f[ POINT_COUNT ];
                newPoints = new CvPoint2D32f[ POINT_COUNT ];
                status = new char[ POINT_COUNT ];

                if ( !image || !eig || !temp || !newImage || !pyramid || !newPyramid || !points || !newPoints || !status) {
                        std::cout << "Nem sikerult memoriat foglalni!" << std::endl;
                }
        }

        cvCvtColor( frame, grayFrame, CV_BGR2GRAY );
        cvResize( grayFrame, image );

}

/**
 *  Transzformacio lekerdezese, utoszures utan.
 *
 *  @return 4 elemu vektor: (xtrans, ytrans, scale, angle)
 */
std::vector<float> WebcamInterface::getMotion( void ) {
        int i, j;
        std::vector<float> avg;
        std::vector<float> weights;
        std::vector<float> lastShift;

        lastShift = getShift( );

        shifts[0] = shifts[1];
        shifts[1] = shifts[2];
        shifts[2] = lastShift;

        avg.push_back( 0 );
        avg.push_back( 0 );
        avg.push_back( 0 );
        avg.push_back( 0 );
        
        weights.push_back( 0.2 );
        weights.push_back( 0.3 );
        weights.push_back( 0.5 );

        for( i = 0; i < 4; i++ )
                for( j = 0; j < MOTION_BUFFER_SIZE; j++ )
                        avg[i] += shifts[j][i] * weights[j];

/*        avg[0] /= MOTION_BUFFER_SIZE;
        avg[1] /= MOTION_BUFFER_SIZE;
        avg[2] /= MOTION_BUFFER_SIZE;
        avg[3] /= MOTION_BUFFER_SIZE;*/

        return avg;
}

/**
 *  0 s N-1 kztti vletlen egsz
 */
static int random( int N ) {
        return  ( rand() % N );
}

/**
 *  0 s n-1 kztti 4db vletlen egzs
 */
static int random4( int n, std::vector<int>& p) {
        int i, j, r;

        std::vector<int> v;

        v.clear( );

        for( i = 0; i < n; i++ )
                v.push_back( i );


        p.clear( );
        for( i = 0; i < 4; i++ ) {
                r = random( n );
                p.push_back( v[r] );
                v[r] = v[n-1];
                n--;
       }
}

#define THRESHOLD ( 10.0 )

/**
 *  Outlier pontmegfeleltetesek kiszurese RANSAC algoritmussal
 *  Az outlierek status-at 0-ra allitja.
 */
void WebcamInterface::RANSAC( int n ) {
        int iter, i;
        std::vector< int > p, bestPoints;
        double bestVotes, votes;

        static Matrix<double> M, A(3, 4), B(3, 4), Ainv, bestM;
        static Vector<double> V1(3);
        static Vector<double> V2(3);
        static Vector<double> V3(3);
        static Vector<double> V4(3);


        /**
         *  pont 4-esek vletlen kivlasztsa, s szavazs
         */
        srand( time( 0 ) );
        bestVotes = -1;
        for( iter = 0; iter < 100; iter++ ) {
                /** Ngy pont vletlenszer kivlasztsa */
                random4( n, p );

                /** Transzformci meghatrozsa */
                for( i = 0; i < 4; i++ ) {
                        A.getElement( 0, i ) = points[p[i]].x;
                        A.getElement( 1, i ) = points[p[i]].y;
                        A.getElement( 2, i ) = 1.0;

                        B.getElement( 0, i ) = newPoints[p[i]].x;
                        B.getElement( 1, i ) = newPoints[p[i]].y;
                        B.getElement( 2, i ) = 1.0;
                }
                SVDInverse( A, Ainv );
                M = B * Ainv;

                /** Szavazs */
                votes = 0;
                for( i = 0; i < n; i++ ) {

                        V1.getElement(0) = points[i].x;
                        V1.getElement(1) = points[i].y;
                        V1.getElement(2) = 1.0;
                        V2.getElement(0) = newPoints[i].x;
                        V2.getElement(1) = newPoints[i].y;
                        V2.getElement(2) = 1.0;

                        V3 = M * V1;

                        V4 = (V2-V3);

                        if ( V4.length() < THRESHOLD ) {
                                votes += 1;//THRESHOLD-V4.length();
                        }
                }
                if ( bestVotes==-1 || votes > bestVotes ) {
                        bestVotes = votes;
                        bestPoints = p;
                        bestM = M;
                }
        }

        int db = 0;

        /** A legjobb pontngyes, s rjuk szavazk sszegyjtse */
        for( i = 0; i < n; i++ ) {
                        V1.getElement(0) = points[i].x;
                        V1.getElement(1) = points[i].y;
                        V1.getElement(2) = 1.0;
                        V2.getElement(0) = newPoints[i].x;
                        V2.getElement(1) = newPoints[i].y;
                        V2.getElement(2) = 1.0;

                        V3 = bestM * V1;

                        V4 = (V2-V3);

                        if ( V4.length() < THRESHOLD ) {
                                status[i] = 1;
                                db++;
                        }else
                                status[i] = 0;
       }

}

/**
 *  Kp histogram kiegyenltse
 */
void WebcamInterface::histogramEqualization( IplImage * img ) {
        double pA[256];
        double T[256];
        int r;
        int i, j, k;
        double sum;
        double imageSize = double(img->height * img->width);
        
        for( i = 0; i < 256; i++ ) pA[i] = 0.0;

        for( i = 0; i < img->height; i++ )
                for( j = 0; j < img->width; j++ )
                        pA[ (unsigned char)(img->imageData[ i*img->width+j ]) ] += 1.0;

        for( i = 0; i < 256; i++ ) pA[i] /= imageSize;

        T[0] = pA[0];
        for( i = 1; i < 256; i++ ) T[i] = T[i-1] + pA[i];

        for( i = 0; i < img->height; i++ )
                for( j = 0; j < img->width; j++ ) {
                        r = (unsigned char)(img->imageData[ i*img->width+j ]);

                        img->imageData[ i*img->width+j ] = (char)((255.0*T[r] ));
                }
}

/**
 *  Uj kep felvetele, transzformacio kiszamitasa.
 *
 *  @return 4 elemu vektor: (xtrans, ytrans, scale, angle)
 */
std::vector<float> WebcamInterface::getShift( void ) {
        int i, j, n;
        std::vector<float> motion;
        int count;

        /** Kezdetben res traf */
        motion.push_back( 0 );
        motion.push_back( 0 );
        motion.push_back( 1 );
        motion.push_back( 0 );

        /** Kp felvtele, szrkernyalatoss konvertlsa */
        if( !cvGrabFrame( cvCapture ) ) {
                std::cout << "Nem sikerult olvasni a frame-t!" << std::endl;
                return motion;
        }
        frame = cvRetrieveFrame( cvCapture );
        cvCvtColor( frame, grayFrame, CV_BGR2GRAY );
        cvResize( grayFrame, newImage );
        histogramEqualization( grayFrame );
        
        /** Feature pont vlaszts */
        count = POINT_COUNT;
        int mindistance = (int)(20.0 * (double)FRAME_WIDTH / 320.0);

        cvGoodFeaturesToTrack( image, eig, temp, points, &count, 0.01, mindistance, 0, 3, 0, 0.03 );
        cvFindCornerSubPix( image, points, count, cvSize(10, 10), cvSize(-1,-1), cvTermCriteria(CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 10, 0.03 ) );

        /** Ha nincs elg pont, megllunk */
        if( count < 10 ) {
                cvCopy( newImage, image, 0 );

                of_flags = 0;

                return motion;
        }

        /** Pontok lekvetse */
        cvCalcOpticalFlowPyrLK( image, newImage, pyramid, newPyramid, points, newPoints, count, cvSize(30, 30 ), 5, status, 0,
                                        cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 20, 0.02), of_flags);
        of_flags |= CV_LKFLOW_PYR_A_READY;

        /** Sikeresen lekvetett pontok szma */
        n = 0;
        for( i = 0; i < count; i++ )
                if ( status[i] )
                        ++n;

        /** Ha van elg lekvetett pont */
        if ( n > 10 ) {
                j = 0;
                for( i = 0; i < count; i++ )
                        if ( status[i] ) {
                                points[j] = points[i];
                        newPoints[j] = newPoints[i];
                                j++;
                        }


                RANSAC( n );

                /** Ha nem maradt elg pont, megllunk */
                j = 0;
                for( i = 0; i < n; i++ )
                        if( status[i] ) j++;
                        
                if( j < 10 )
                        return motion;


                /** Transzformcis mtrix kiszmtsa */
                Matrix<double> M, A(3, n), B(3, n), Ainv;

                j = 0;
                for( i = 0; i < n; i++ ) {
                        if( !status[i] ) continue;
                        
                        A.getElement( 0, j ) = points[i].x;
                        A.getElement( 1, j ) = points[i].y;
                        A.getElement( 2, j ) = 1.0;

                        B.getElement( 0, j ) = newPoints[i].x;
                        B.getElement( 1, j ) = newPoints[i].y;
                        B.getElement( 2, j ) = 1.0;
                        j++;
                        
                }
                SVDInverse( A, Ainv );

                M = B * Ainv;


                /** Transzformcis paramterek kinyerse */
                double a = M.getElement( 0, 0 );
                double b = M.getElement( 0, 1 );
                double c = M.getElement( 1, 0 );
                double d = M.getElement( 1, 1 );

                double sx = sqrt( a*a + c*c );
                double sy = sqrt( b*b + d*d );
                double s = (sx+sy)/2.0;//sqrt(sx*sx*sy*sy);
                double angle = asin( -c/sx );
                
                motion[0] = M.getElement( 0, 2 );
                motion[1] = M.getElement( 1, 2 );

                motion[2] = 1.0/s;
                motion[3] = angle;
        }

        cvCopy( newImage, image, 0 );
        cvCopy( newPyramid, pyramid, 0 );

        return motion;
}

IplImage * WebcamInterface::getImage( void ) {
        return frame;
}
